import json
from django.test import TestCase
from django.db import models

from flowr.models import (Rule, RuleSet, Flow, State, DCCGraph, Node,
    BaseNodeData)
from flowr import sample_rules as rules

# ============================================================================

#def pprint(data):
#    print(json.dumps(data, sort_keys=True, indent=4, separators=(',', ': ')))

# ============================================================================
# Test Class
# ============================================================================

class Label(BaseNodeData):
    name = models.CharField(max_length=5)

    def __str__(self):
        return 'Label(id=%s %s)' % (self.id, self.name)


class GraphTests(TestCase):
    def setUp(self):
        # create a bunch of test graphs

        # ---
        #    A.A
        #     |
        #    A.B
        self.graph_a = DCCGraph.factory(Label, name='A.A')
        self.graph_a.root.add_child(name='A.B')

        # verify
        self.assertEqual(Node.objects.filter(graph=self.graph_a).count(), 2)
        a, b = self._graph_query(self.graph_a, ['A.A', 'A.B'])
        self.assertEqual(a.parents.count(), 0)
        self.assertEqual(list(a.children.all()), [b])
        self.assertEqual(list(b.parents.all()), [a])
        self.assertEqual(b.children.count(), 0)

        # ---
        #      B.A
        #     /   \
        #   B.B   B.C
        #           \
        #           B.A (cycle)
        self.graph_b = DCCGraph.factory(Label, name='B.A')
        self.graph_b.root.add_child(name='B.B')
        node = self.graph_b.root.add_child(name='B.C')
        node.connect_child(self.graph_b.root)

        # verify
        self.assertEqual(Node.objects.filter(graph=self.graph_b).count(), 3)
        a,b,c = self._graph_query(self.graph_b, ['B.A', 'B.B', 'B.C'])

        self.assertEqual(a, self.graph_b.root)
        self.assertEqual(list(a.parents.all()), [c])
        self.assertEqual(set(a.children.all()), set([b, c]))
        self.assertEqual(list(b.parents.all()), [a])
        self.assertEqual(b.children.count(), 0)
        self.assertEqual(list(c.parents.all()), [a])
        self.assertEqual(list(c.children.all()), [a])

        # ---
        #           C.A
        #          /   \
        #        C.B   C.C
        #             /   \
        #            /     \
        #           /       \
        #          /         \
        #         /           \
        #       C.D           C.E
        #      /   \            \
        #    C.G   C.H          C.F
        #    /       \             \
        #  C.I        C.A (cycle)  C.A (cycle)
        #    \
        #    C.H (loop back)

        self.graph_c = DCCGraph.factory_from_graph(Label, 
            {'name':'C.A'}, [
                ({'name':'C.B'}, []),
                ({'name':'C.C'}, [
                    ({'name':'C.D'}, [
                        ({'name':'C.G'}, [
                            ({'name':'C.I'}, []),
                        ]),
                        ({'name':'C.H'}, []),
                    ]),
                    ({'name':'C.E'}, [
                        ({'name':'C.F'}, []),
                    ]),
                ]),
            ])
        i = self.graph_c.find_nodes(name='C.I').first()
        h = self.graph_c.find_nodes(name='C.H').first()
        i.connect_child(h)
        h.connect_child(self.graph_c.root)
        f = self.graph_c.find_nodes(name='C.F').first()
        f.connect_child(self.graph_c.root)

        # verify
        self.assertEqual(Node.objects.filter(graph=self.graph_c).count(), 9)
        a,b,c,d,e,f,g,h,i = self._graph_query(self.graph_c, ['C.A', 'C.B',
            'C.C', 'C.D', 'C.E', 'C.F', 'C.G', 'C.H', 'C.I'])

        self.assertEqual(a, self.graph_c.root)
        self.assertEqual(set(a.parents.all()), set([f, h]))
        self.assertEqual(set(a.children.all()), set([b, c]))
        self.assertEqual(list(b.parents.all()), [a])
        self.assertEqual(b.children.count(), 0)

        self.assertEqual(list(c.parents.all()), [a])
        self.assertEqual(set(c.children.all()), set([d, e]))

        self.assertEqual(list(d.parents.all()), [c])
        self.assertEqual(set(d.children.all()), set([g, h]))

        self.assertEqual(list(e.parents.all()), [c])
        self.assertEqual(set(e.children.all()), set([f]))

        self.assertEqual(list(f.parents.all()), [e])
        self.assertEqual(set(f.children.all()), set([a]))

        self.assertEqual(list(g.parents.all()), [d])
        self.assertEqual(set(g.children.all()), set([i]))

        self.assertEqual(list(h.parents.all()), [d, i])
        self.assertEqual(set(h.children.all()), set([a]))

        self.assertEqual(list(i.parents.all()), [g])
        self.assertEqual(set(i.children.all()), set([h]))

    def _graph_query(self, graph, values):
        nodes = []
        for value in values:
            nodes.append(graph.find_nodes(name=value).get())

        return nodes

    def _c_graph_nodes(self):
        return self._graph_query(self.graph_c, ['C.A', 'C.B', 'C.C', 'C.D', 
            'C.E', 'C.F', 'C.G', 'C.H', 'C.I'])

    def test_ancestors(self):
        a,b,c,d,e,f,g,h,i = self._c_graph_nodes()

        self.assertEqual(set(a.ancestors()), set([f, e, c, h, d, i, g]))
        self.assertEqual(set(b.ancestors()), set([a, f, e, c, h, d, i, g]))
        self.assertEqual(set(c.ancestors()), set([a, f, e, h, d, i, g]))
        self.assertEqual(set(d.ancestors()), set([c, a, f, e, h, i, g]))
        self.assertEqual(set(e.ancestors()), set([c, a, f, h, d, i, g]))
        self.assertEqual(set(f.ancestors()), set([e, c, a, h, d, i, g]))
        self.assertEqual(set(g.ancestors()), set([d, c, a, f, e, h, i]))
        self.assertEqual(set(h.ancestors()), set([d, c, a, f, e, i, g]))
        self.assertEqual(set(i.ancestors()), set([g, d, c, a, f, e, h,]))

        self.assertEqual(set(a.ancestors_root()), set([]))
        self.assertEqual(set(b.ancestors_root()), set([a]))
        self.assertEqual(set(c.ancestors_root()), set([a]))
        self.assertEqual(set(d.ancestors_root()), set([c, a]))
        self.assertEqual(set(e.ancestors_root()), set([c, a]))
        self.assertEqual(set(f.ancestors_root()), set([e, c, a]))
        self.assertEqual(set(g.ancestors_root()), set([d, c, a]))
        self.assertEqual(set(h.ancestors_root()), set([d, c, a, i, g]))
        self.assertEqual(set(i.ancestors_root()), set([g, d, c, a]))

    def test_descendents(self):
        a,b,c,d,e,f,g,h,i = self._c_graph_nodes()

        self.assertEqual(set(a.descendents()), set([b, c, d, e, f, g, h, i]))
        self.assertEqual(set(b.descendents()), set([]))
        self.assertEqual(set(c.descendents()), set([a, b, d, e, f, g, h, i]))
        self.assertEqual(set(d.descendents()), set([a, b, c, e, f, g, h, i]))
        self.assertEqual(set(e.descendents()), set([a, b, c, d, f, g, h, i]))
        self.assertEqual(set(f.descendents()), set([a, b, c, d, e, g, h, i]))
        self.assertEqual(set(g.descendents()), set([a, b, c, d, e, f, h, i]))
        self.assertEqual(set(h.descendents()), set([a, b, c, d, e, f, g, i]))
        self.assertEqual(set(i.descendents()), set([a, b, c, d, e, f, g, h]))

        self.assertEqual(set(a.descendents_root()), 
            set([b, c, d, e, f, g, h, i]))
        self.assertEqual(set(b.descendents_root()), set([]))
        self.assertEqual(set(c.descendents_root()), set([d, e, f, g, h, i, a]))
        self.assertEqual(set(d.descendents_root()), set([g, h, i, a]))
        self.assertEqual(set(e.descendents_root()), set([f, a]))
        self.assertEqual(set(f.descendents_root()), set([a]))
        self.assertEqual(set(g.descendents_root()), set([i, h, a]))
        self.assertEqual(set(h.descendents_root()), set([a]))
        self.assertEqual(set(i.descendents_root()), set([h, a]))

    def test_cytoscape(self):
        a,b,c,d,e,f,g,h,i = self._c_graph_nodes()

        extra_fields = lambda n: {'name':n.data.name}
        result = json.loads(self.graph_c.cytoscape_json(extra_fields))
        nodes = result['nodes']
        edges = result['edges']
        self.assertEqual(len(nodes), 9)
        self.assertEqual(len(edges), 11)

        for node in [a,b,c,d,e,f,g,h,i]:
            expected = {
                'data':{
                    'id':'n%s' % node.id,
                    'name':node.data.name,
                }
            }
            self.assertIn(expected, nodes)

        expected_edges = [(a,b), (a,c), (c,d), (c,e), (d, g), (d, h), (e, f),
            (f, a), (g, i), (h, a), (i, h)]
        for edge in expected_edges:
            expected = {
                'data' : { 
                    'id':'e%s_%s' % (edge[0].id, edge[1].id), 
                    'source':'n%s' % edge[0].id, 
                    'target':'n%s' % edge[1].id,
                }
            }
            self.assertIn(expected, edges)

    def test_can_delete(self):
        a,b,c,d,e,f,g,h,i = self._c_graph_nodes()

        self.assertFalse( a.can_remove() )
        self.assertFalse( c.can_remove() )
        self.assertFalse( d.can_remove() )
        self.assertFalse( e.can_remove() )
        self.assertFalse( g.can_remove() )
        self.assertFalse( i.can_remove() )

        self.assertTrue( b.can_remove() )
        self.assertTrue( f.can_remove() )
        self.assertTrue( h.can_remove() )

    def test_delete(self):
        a,b,c,d,e,f,g,h,i = self._c_graph_nodes()

        with self.assertRaises(AttributeError):
            a.remove()

        expected = b.data
        result = b.remove()
        self.assertEqual(expected, result)
        self.assertEqual(Node.objects.filter(graph=self.graph_c).count(), 8)
        self.assertEqual(set(a.descendents()), set([c, d, e, f, g, h, i]))

        expected = h.data
        result = h.remove()
        self.assertEqual(expected, result)
        self.assertEqual(Node.objects.filter(graph=self.graph_c).count(), 7)
        self.assertEqual(set(a.descendents()), set([c, d, e, f, g, i]))
        self.assertEqual(i.children.count(), 0)
        self.assertEqual(set(d.descendents_root()), set([g, i]))

    def test_prune(self):
        a,b,c,d,e,f,g,h,i = self._c_graph_nodes()

        # -- simple prune of childless node
        expected = [b, ]
        result = b.prune_list()
        self.assertEqual(expected, result)

        expected = [b.data, ]
        result = b.prune()
        self.assertEqual(expected, result)
        self.assertEqual(set(a.descendents()), set([c, d, e, f, g, h, i]))
        self.assertEqual(a.children.count(), 1)
        self.assertEqual(Node.objects.filter(graph=self.graph_c).count(), 8)

        # -- more complex prune of node with children and cycles
        expected = [d, g, h, i]
        result = d.prune_list()
        self.assertEqual(set(expected), set(result))

        expected = [d.data, g.data, h.data, i.data]
        result = d.prune()
        self.assertEqual(set(expected), set(result))
        self.assertEqual(set(a.descendents()), set([c, e, f]))
        self.assertEqual(c.children.count(), 1)
        self.assertEqual(Node.objects.filter(graph=self.graph_c).count(), 4)

    def test_errors(self):
        # Check we can't connect nodes from different graphs
        a = self._graph_query(self.graph_a, ['A.A', ])[0]
        b = self._graph_query(self.graph_a, ['B.A', ])[0]
        
        with self.assertRaises(AttributeError):
            a.connect_child(b)

        # check we can't create a graph without a DataNode extender
        with self.assertRaises(AttributeError):
            DCCGraph.factory(object)

    def test_coverage(self):
        # misc items to get full coverage
        str(self.graph_a)
        str(self.graph_a.root)
        str(self.graph_a.root.data)

# ============================================================================

class FlowTests(TestCase):
    def setUp(self):
        self.rule_set = RuleSet.factory('My Rules', rules.A)

    def test_coverage(self):
        """Miscellaneous stuff to get our coverage to 100%"""
        Rule.on_enter(None)
        Rule.on_leave(None)
        with self.assertRaises(NotImplementedError):
            Rule.edit_screen(None, None)

        str(self.rule_set)
        rules.A.display_name()

    def test_rule_cytoscape(self):
        result = json.loads(self.rule_set.cytoscape_json())
        nodes = result['nodes']
        edges = result['edges']
        self.assertEqual(len(nodes), 5)
        self.assertEqual(len(edges), 5)

        for rule in [rules.A, rules.B, rules.C, rules.D, rules.E]:
            expected = {
                'data':{
                    'id':rule.name,
                    'label':rule.name,
                }
            }
            self.assertIn(expected, nodes)

        expected_edges = [(rules.A, rules.B), (rules.A, rules.C), 
            (rules.C, rules.D), (rules.C, rules.E), (rules.E, rules.A)]
        for edge in expected_edges:
            expected = {
                'data' : { 
                    'id':'%s_%s' % (edge[0].name, edge[1].name), 
                    'source':edge[0].name, 
                    'target':edge[1].name,
                }
            }
            self.assertIn(expected, edges)

    def test_flow1(self):
        """Tests a simple flow:
                    A
                    |
                    B
        """
        flow = Flow.factory('Flow', self.rule_set)
        self.assertFalse(flow.in_use())
        root = flow.root_data

        # should not be able to add D as it is against the rules
        with self.assertRaises(AttributeError):
            root.add_child_rule(rules.D)

        # should be able to add B
        root.add_child_rule(rules.B)

        # should not be able to add B again as A isn't multipath
        with self.assertRaises(AttributeError):
            root.add_child_rule(rules.B)

        # --- State testing
        rules.done_enter = []
        rules.done_leave = []

        # starting should call A's on_enter
        state = State.start(flow)
        self.assertEqual(rules.done_enter, ['A'])
        self.assertEqual(rules.done_leave, [])

        # next step
        state.next_state()
        self.assertEqual(rules.done_leave, ['A'])
        self.assertEqual(rules.done_enter, ['A', 'B'])

        # no next step in Flow
        with self.assertRaises(AttributeError):
            state.next_state()

        # force __str__ calls for coverage
        str(flow)
        str(root)
        str(state)

    def test_flow2(self):
        """Tests multipath, looping flow:
                   A
                   |
                   C
                  / \
                 D   E
                      \
                       A (loops)
        """
        flow = Flow.factory(name='Flow', rule_set=self.rule_set)
        root = flow.root_data
        node = root.add_child_rule(rules.C)
        node.add_child_rule(rules.D)

        # should not be able to add D a second time
        with self.assertRaises(AttributeError):
            node.add_child_rule(rules.D)

        node = node.add_child_rule(rules.E)
        node.connect_child(root)

        # -- start testing the flow
        rules.done_enter = []
        rules.done_leave = []
        state = State.start(flow=flow)
        self.assertEqual(rules.done_enter, ['A'])
        self.assertEqual(rules.done_leave, [])

        state.next_state()
        self.assertEqual(rules.done_enter, ['A', 'C'])
        self.assertEqual(rules.done_leave, ['A'])

        # next step requires a choice, can't use parameterless call
        with self.assertRaises(AttributeError):
            state.next_state()

        # next step must be D or E
        with self.assertRaises(AttributeError):
            state.next_state(rules.A)

        state.next_state(rules.E)
        state.next_state(rules.A)
        self.assertEqual(rules.done_enter, ['A', 'C', 'E', 'A'])
        self.assertEqual(rules.done_leave, ['A', 'C', 'E'])

    def test_delete(self):
        flow = Flow.factory(name='Flow', rule_set=self.rule_set)
        flow.delete()

        self.assertEqual(Flow.objects.count(), 0)
        self.assertEqual(DCCGraph.objects.count(), 0)



# flowr.models.py
import logging, json

from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.db.models.signals import post_delete
from django.dispatch import receiver

logger = logging.getLogger(__name__)

# ============================================================================
# Directed, Connected(ish), Cyclic Graph

class DCCGraph(models.Model):
    """``DCCGraph`` is a modified Directed, Connected, Cyclic Graph.  It
    consists of a series of nodes connected in a directed fashion with a
    defined root.  All nodes in the graph are reachable by traversing the
    edges from the root.  This is similar to a DAG with a single root, except
    that cycles are allowed.

    Each :class:`Node` in the ``DCCGraph`` may have a related data class in
    order to associated arbitrary with the node.  A data class can be any
    django model, that inherits from :class:`BaseNodeData`, the type of which
    is specified upon the creation of the ``DCCGraph``.
    """
    data_content_type = models.ForeignKey(ContentType)
    root = models.ForeignKey('flowr.Node', blank=True, null=True)

    class Meta:
        verbose_name = 'DCC Graph'

    def __str__(self):
        return 'DCCGraph(id=%s)' % self.id

    @classmethod
    @transaction.atomic
    def factory(cls, data_class, **kwargs):
        """Creates a ``DCCGraph``, a root :class:`Node`: and the node's
        associated data class instance.  This factory is used to get around
        the chicken-and-egg problem of the ``DCCGraph`` and ``Nodes`` within
        it having pointers to each other.

        :param data_class: 
            django model class that extends :class:`BaseNodeData` and is used
            to associate information with the nodes in this graph
        :param kwargs: 
            arguments to pass to constructor of the data class instance that
            is to be created for the root node
        :returns: 
            instance of the newly created ``DCCGraph``
        """
        if not issubclass(data_class, BaseNodeData):
            raise AttributeError('data_class must be a BaseNodeData extender')

        content_type = ContentType.objects.get_for_model(data_class)
        graph = DCCGraph.objects.create(data_content_type=content_type)
        node = Node.objects.create(graph=graph)
        data_class.objects.create(node=node, **kwargs)
        graph.root = node
        graph.save()
        return graph

    @classmethod
    def _depth_create(cls, node, child_args, grandchildren):
        child = node.add_child(**child_args)
        for grandkid in grandchildren:
            cls._depth_create(child, grandkid[0], grandkid[1])

    @classmethod
    @transaction.atomic
    def factory_from_graph(cls, data_class, root_args, children):
        """Creates a ``DCCGraph`` and corresponding nodes.  The root_args parm
        is a dictionary specifying the parameters for creating a :class:`Node`
        and its corresponding :class:`BaseNodeData` subclass from the
        data_class specified.  The children parm is an iterable containing
        pairs of dictionaries and iterables, where the dictionaries specify
        the parameters for a :class:`BaseNodeData` subclass and the iterable
        the list of children.

        Example::

            DCCGraph.factory_from_graph(Label, 
                {'name':'A'}, [
                    ({'name':'B', []),
                    ({'name':'C', [])
                ])

        creates the graph::

                 A
                / \
               B   C

        :param data_class: 
            django model class that extends :class:`BaseNodeData` and is used
            to associate information with the Nodes in this graph
        :param root_args: 
            dictionary of arguments to pass to constructor of the data class
            instance that is to be created for the root node
        :param children: 
            iterable with a list of dictionary and iterable pairs
        :returns: 
            instance of the newly created ``DCCGraph``
        """
        graph = cls.factory(data_class, **root_args)
        for child in children:
            cls._depth_create(graph.root, child[0], child[1])

        return graph

    def find_nodes(self, **kwargs):
        """Searches the data nodes that are associated with this graph using
        the key word arguments as a filter and returns a
        :class:`django.db.models.query.QuerySet`` of the attached
        :class:`Node` objects.

        :param kwargs: 
            filter arguments applied to searching the :class:`BaseNodeData`
            subclass associated with this graph.
        :returns: 
            ``QuerySet`` of :class:`Node` objects 
        """
        filter_args = {}
        classname = self.data_content_type.model_class().__name__.lower()
        for key, value in kwargs.items():
            filter_args['%s__%s' % (classname, key)] = value

        return Node.objects.filter(**filter_args)

    def cytoscape_json(self, extra_fields=lambda x:{}):
        data = {
            'nodes':[],
            'edges':[],
        }
        for node in Node.objects.filter(graph=self):
            node_data = {
                'data':{ 
                    'id':'n%s' % node.id,
                }
            }
            for key, value in extra_fields(node).items():
                node_data['data'][key] = value

            data['nodes'].append(node_data)

            for child in node.children.all():
                data['edges'].append({ 
                    'data':{ 
                        'id':'e%s_%s' % (node.id, child.id),
                        'source':'n%s' % node.id,
                        'target':'n%s' % child.id,
                    }
                })

        return json.dumps(data)


class BaseNodeData(models.Model):
    """Subclasses of this object are used to store information associated with
    a :class:`Node` in the :class:`DCCGraph`."""
    node = models.OneToOneField('flowr.Node')

    class Meta:
        abstract = True


class Node(models.Model):
    """Represents a single ``Node`` in the :class:`DCCGraph`.  Can be
    associated with custom data using a subclass of :class:`BaseNodeData`."""
    graph = models.ForeignKey(DCCGraph)
    parents = models.ManyToManyField('self', symmetrical=False,
        related_name='parent_nodes')
    children = models.ManyToManyField('self', symmetrical=False,
        related_name='child_nodes')

    def __str__(self):
        return 'Node(id=%s)' % self.id

    @property
    def data(self):
        data_model = self.graph.data_content_type.model_class()
        return data_model.objects.get(node=self)

    def is_root(self):
        return self == self.graph.root

    def add_child(self, **kwargs):
        """Creates a new ``Node`` based on the extending class and adds it as
        a child to this ``Node``.

        :param kwargs: 
            arguments for constructing the data object associated with this
            ``Node``
        :returns: 
            extender of the ``Node`` class
        """
        data_class = self.graph.data_content_type.model_class()
        node = Node.objects.create(graph=self.graph)
        data_class.objects.create(node=node, **kwargs)
        node.parents.add(self)
        self.children.add(node)
        return node

    def connect_child(self, node):
        """Adds the given node as a child to this one.  No new nodes are
        created, only connections are made.
        
        :param node:
            a ``Node`` object to connect
        """
        if node.graph != self.graph:
            raise AttributeError('cannot connect nodes from different graphs')

        node.parents.add(self)
        self.children.add(node)

    def _depth_ascend(self, node, visited, limit=False):
        for parent in node.parents.all():
            if limit and parent.is_root():
                visited.add(parent)

            if parent not in visited:
                visited.add(parent)
                self._depth_ascend(parent, visited, limit)

    def ancestors(self):
        """Returns a list of the ancestors of this node."""
        ancestors = set([])
        self._depth_ascend(self, ancestors)
        try:
            ancestors.remove(self)
        except KeyError:
            # we weren't ancestor of ourself, that's ok
            pass

        return list(ancestors)

    def ancestors_root(self):
        """Returns a list of the ancestors of this node but does not pass the
        root node, even if the root has parents due to cycles."""
        if self.is_root():
            return []

        ancestors = set([])
        self._depth_ascend(self, ancestors, True)
        try:
            ancestors.remove(self)
        except KeyError:
            # we weren't ancestor of ourself, that's ok
            pass

        return list(ancestors)

    def _depth_descend(self, node, visited, limit=False):
        for child in node.children.all():
            if limit and child.is_root():
                visited.add(child)

            if child not in visited:
                visited.add(child)
                self._depth_descend(child, visited, limit)

    def descendents(self):
        """Returns a list of descendents of this node."""
        visited = set([])
        self._depth_descend(self, visited)
        try:
            visited.remove(self)
        except KeyError:
            # we weren't descendent of ourself, that's ok
            pass

        return list(visited)

    def descendents_root(self):
        """Returns a list of descendents of this node, if the root node is in
        the list (due to a cycle) it will be included but will not pass
        through it.  
        """
        visited = set([])
        self._depth_descend(self, visited, True)
        try:
            visited.remove(self)
        except KeyError:
            # we weren't descendent of ourself, that's ok
            pass

        return list(visited)

    def can_remove(self):
        """Returns True if it is legal to remove this node and still leave the
        graph as a single connected entity, not splitting it into a forest.
        Only nodes with no children or those who cause a cycle can be deleted.
        """
        if self.children.count() == 0:
            return True

        ancestors = set(self.ancestors_root())
        children = set(self.children.all())
        return children.issubset(ancestors)

    def remove(self):
        """Removes the node from the graph.  Note this does not remove the
        associated data object.  See :func:`Node.can_remove` for limitations
        on what can be deleted.

        :returns: 
            :class:`BaseNodeData` subclass associated with the deleted Node

        :raises AttributeError:
           if called on a ``Node`` that cannot be deleted
        """
        if not self.can_remove():
            raise AttributeError('this node cannot be deleted')

        data = self.data
        self.parents.remove(self)
        self.delete()
        return data

    def prune(self):
        """Removes the node and all descendents without looping back past the
        root.  Note this does not remove the associated data objects.

        :returns:
            list of :class:`BaseDataNode` subclassers associated with the
            removed ``Node`` objects.
        """
        targets = self.descendents_root()
        try:
            targets.remove(self.graph.root)
        except ValueError:
            # root wasn't in the target list, no problem
            pass

        results = [n.data for n in targets]
        results.append(self.data)
        for node in targets:
            node.delete()

        for parent in self.parents.all():
            parent.children.remove(self)

        self.delete()
        return results

    def prune_list(self):
        """Returns a list of nodes that would be removed if prune were called
        on this element.
        """
        targets = self.descendents_root()
        try:
            targets.remove(self.graph.root)
        except ValueError:
            # root wasn't in the target list, no problem
            pass

        targets.append(self)
        return targets

# ============================================================================

class TimeTrackedModel(models.Model):
    """Abstract model for create & update fields.  """

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True

# ============================================================================
# Rule Models
# ============================================================================

class _MetaRule(type):
    def __new__(cls, name, bases, classdict):
        klass = type.__new__(cls, name, bases, dict(classdict))
        klass.class_label = '%s.%s' % (klass.__module__, klass.__name__)
        klass.name = klass.__name__
        return klass

class Rule(metaclass=_MetaRule):
    """Subclasses of ``Rule`` are associated with states in the flow graph.
    As states are entered and exited, the :func:`Rule.on_enter` and 
    :func:`Rule.on_leave` methods are called respectively.

    :param children: 
        list of other subclasses of ``Rule`` that are allowed as children of
        this node in the flow.  Cycles are allowed even with the ``Rule``
        being its own child.

    :param multiple_paths: 
        ``True`` if the exit of this state can have multiple choices.
        Multiple children mean you can construct different exits in the flow,
        multiple paths mean there is a branching choice in the flow that can
        go down the path of more than one of the children.  Defaults to
        ``False``.

    :param has_edit_screen: 
        ``True`` if the GUI should show an edit screen when a node for this
        ``Rule`` is selected.   If ``True``, the :func:`Rule.edit_screen`
        method will be called when the user clicks on an associated node.
        Defaults to ``False``.
    """
    children = []
    multiple_paths = False
    has_edit_screen = False

    @classmethod
    def on_enter(cls, state):
        """Called when a state corresponding to this ``Rule`` is entered."""
        pass

    @classmethod
    def on_leave(cls, state):
        """Called when a state corresponding to this ``Rule`` is left."""
        pass

    @classmethod
    def edit_screen(cls, request, flow_node_data):
        """Called if ``Rule.has_edit_screen`` is ``True`` and the user clicks
        on a flow node associated with this ``Rule``.

        :returns: 
            implementation should return the HTML to display when the user
            selects this node on the edit screen.
        """
        raise NotImplementedError()


class RuleSet(TimeTrackedModel):
    """Contains the name of a :class:`Rule` subclass that acts as the root in
    a rule graph.  All :class:`Flow` objects are associated with a ``RuleSet``
    and the state flow graphs of the flow are a subset of those defined in the
    :class:`Rule` objects.
    """
    name = models.CharField(max_length=80)
    root_rule_label = models.CharField(max_length=80)

    class Meta:
        verbose_name = 'Rule Set'

    def __init__(self, *args, **kwargs):
        super(RuleSet, self).__init__(*args, **kwargs)

        # ---
        # define:
        #   root_rule
        #
        # as parameters on class
        if self.root_rule_label:
            (modname, classname) = self.root_rule_label.rsplit('.', 1)
            mod = __import__(modname, globals(), locals(), [classname])
            self.root_rule = getattr(mod, classname)
            self.root_rule_name = self.root_rule.__name__

    def __str__(self):
        return 'RuleSet(id=%s, %s:%s)' % (self.id, self.name,
            self.root_rule_label)

    @classmethod
    def factory(self, name, root_rule):
        return RuleSet.objects.create(name=name,
            root_rule_label=root_rule.class_label)


    def _depth_cyto_traverse(self, rule, nodes, edges, visited):
        if rule not in visited:
            visited.add(rule)

            nodes.append({
                'data':{ 
                    'id':rule.name,
                    'label':rule.name,
                }
            })

            for child in rule.children:
                edges.append({ 
                    'data':{ 
                        'id':'%s_%s' % (rule.name, child.name),
                        'source':rule.name,
                        'target':child.name,
                    }
                })
                self._depth_cyto_traverse(child, nodes, edges, visited)

    def cytoscape_json(self, extra_fields=lambda x:{}):
        nodes = []
        edges = []
        visited = set([])

        self._depth_cyto_traverse(self.root_rule, nodes, edges, visited)

        data = {
            'nodes':nodes,
            'edges':edges,
        }

        return json.dumps(data)

# ============================================================================
# Flow Models
# ============================================================================

class FlowNodeData(TimeTrackedModel, BaseNodeData):
    rule_label = models.CharField(max_length=80)

    class Meta:
        verbose_name = 'Flow Node Data'
        verbose_name_plural = 'Flow Node Data'

    def __init__(self, *args, **kwargs):
        super(FlowNodeData, self).__init__(*args, **kwargs)

        # ---
        # define:
        #   rule
        #   flow
        #
        # as parameters on object
        if self.rule_label:
            (modname, classname) = self.rule_label.rsplit('.', 1)
            mod = __import__(modname, globals(), locals(), [classname])
            self.rule = getattr(mod, classname)
            self.rule_name = self.rule.__name__

        if self.node:
            try:
                self.flow = Flow.objects.get(state_graph=self.node.graph)
            except Flow.DoesNotExist:
                # first time a Flow is created there is a chicken-and-egg
                # problem with the Flow and its first Node, ignore it
                self.flow = None

    def __str__(self):
        return 'FlowNodeData(id=%s, %s)' % (self.id, self.rule_label)

    def _child_allowed(self, child_rule):
        """Called to verify that the given rule can become a child of the
        current node.  

        :raises AttributeError: 
            if the child is not allowed
        """
        num_kids = self.node.children.count()
        num_kids_allowed = len(self.rule.children)
        if not self.rule.multiple_paths:
            num_kids_allowed = 1

        if num_kids >= num_kids_allowed:
            raise AttributeError('Rule %s only allows %s children' % (
                self.rule_name, self.num_kids_allowed))

        # verify not a duplicate
        for node in self.node.children.all():
            if node.data.rule_label == child_rule.class_label:
                raise AttributeError('Child rule already exists')

        # check if the given rule is allowed as a child
        if child_rule not in self.rule.children:
            raise AttributeError('Rule %s is not a valid child of Rule %s' % (
                child_rule.__name__, self.rule_name))

    def add_child_rule(self, child_rule):
        """Add a child path in the :class:`Flow` graph using the given 
        :class:`Rule` subclass.  This will create a new child :class:`Node` in
        the associated :class:`Flow` object's state graph with a new
        :class:`FlowNodeData` instance attached.
        
        The :class:`Rule` must be allowed at this stage of the flow according
        to the hierarchy of rules.

        :param child_rule: 
            :class:`Rule` class to add to the flow as a child of :class:`Node`
            that this object owns
        :returns: 
            ``FlowNodeData`` that was added
        """
        self._child_allowed(child_rule)
        child_node = self.node.add_child(rule_label=child_rule.class_label)
        return child_node.data

    def connect_child(self, child_node):
        """Adds a connection to an existing rule in the :class`Flow` graph.
        The given :class`Rule` subclass must be allowed to be connected at
        this stage of the flow according to the hierarchy of rules.

        :param child_node: 
            ``FlowNodeData`` to attach as a child
        """
        self._child_allowed(child_node.rule)
        self.node.connect_child(child_node.node)


class Flow(TimeTrackedModel):
    name = models.CharField(max_length=25)
    rule_set = models.ForeignKey(RuleSet)
    state_graph = models.ForeignKey(DCCGraph)

    def __str__(self):
        return 'Flow(id=%s %s)' % (self.id, self.name)

    @classmethod
    def factory(cls, name, rule_set):
        graph = DCCGraph.factory(FlowNodeData, 
            rule_label=rule_set.root_rule_label)
        flow = Flow.objects.create(name=name, rule_set=rule_set,
            state_graph=graph)
        return flow

    def in_use(self):
        """Returns True if there is a :class:`State` object that uses this
        ``Flow``"""
        state = State.objects.filter(flow=self).first()
        return bool(state)

    @property
    def root_data(self):
        """Returns the :class:`FlowDataNode` object for the root node in the
        graph."""
        return self.state_graph.root.data


@receiver(post_delete, sender=Flow)
def _flow_post_delete(sender, **kwargs):
    # Flow creates a DCCGraph instance, this should be removed when Flow is
    # deleted
    kwargs['instance'].state_graph.delete()


class State(TimeTrackedModel):
    flow = models.ForeignKey(Flow)
    current_node = models.ForeignKey(Node)

    @classmethod
    def start(self, flow):
        """Factory method for a running state based on a flow.  Creates and
        returns a ``State`` object and calls the associated
        :func:`Rule.on_enter` method.

        :param flow: 
            :class:`Flow` which defines this state machine
        :returns: 
            newly created instance
        """
        state = State.objects.create(flow=flow, 
            current_node=flow.state_graph.root)
        flow.state_graph.root.data.rule.on_enter(state)
        return state

    def __str__(self):
        return 'State(id=%s current=%s)' % (self.id,
            self.current_node.data.rule_name)

    def next_state(self, rule=None):
        """Proceeds to the next step in the flow.  Calls the associated
        :func:`Rule.on_leave` method for the for the current rule and the
        :func:`Rule.on_enter` for the rule being entered.  If the current step
        in the flow is multipath then a valid :class:`Rule` subclass must be
        passed into this call.  

        If there is only one possible path in the flow and a :class:`Rule` is
        given it will be ignored.

        :param rule: 
            if the current :class:`Rule` in the :class:`Flow` is multipath
            then the next :class:`Rule` in the flow must be provided.
        """
        num_kids = self.current_node.children.count()
        next_node = None
        if num_kids == 0:
            raise AttributeError('No next state in this Flow id=%s' % (
                self.flow.id))
        elif num_kids == 1:
            next_node = self.current_node.children.first()
        else:
            if not rule:
                raise AttributeError(('Current Rule %s is multipath but no '
                    'choice was passed in') % self.current_node.data.rule_name)

            for node in self.current_node.children.all():
                if node.data.rule_label == rule.class_label:
                    next_node = node
                    break

            if not next_node:
                raise AttributeError(('Current Rule %s is multipath and the '
                    'Rule choice passed in was not in the Flow') % (
                    self.current_node.data.rule_name))

        self.current_node.data.rule.on_leave(self)
        next_node.data.rule.on_enter(self)
        self.current_node = next_node
        self.save()



from django.conf.urls import patterns, url

urlpatterns = patterns('flowr.views',
    url(r'view_rule_set_tree/(\d+)/$', 'view_rule_set_tree',
        name='flowr-view-rule-set-tree'),
    url(r'create_flow/(\d+)/$', 'create_flow', name='flowr-create-flow'),
    url(r'edit_flow/(\d+)/$', 'edit_flow', name='flowr-edit-flow'),

    url(r'node_selected/n(\d+)/$', 'node_selected'),
    url(r'node_remove/n(\d+)/$', 'node_remove'),
    url(r'node_prune/n(\d+)/$', 'node_prune'),
)



# flowr.admin.py

import logging

from django.contrib import admin
from django.core.urlresolvers import reverse

from flowr.models import DCCGraph, Node, RuleSet, Flow, FlowNodeData, State

logger = logging.getLogger(__name__)

# =============================================================================
# Graph Admin Classes
# =============================================================================

@admin.register(DCCGraph)
class DCCGraphAdmin(admin.ModelAdmin):
    list_display = ('id', 'self_model', 'self_root', 'self_num_nodes', 
        'self_nodes')

    def self_model(self, obj):
        return '%s' % obj.data_content_type.model_class().__name__
    self_model.short_description = 'Data Model'

    def self_root(self, obj):
        return '%s' % obj.root
    self_root.short_description = 'Root'

    def self_num_nodes(self, obj):
        return Node.objects.filter(graph=obj).count()
    self_num_nodes.short_description = '# Nodes'

    def self_nodes(self, obj):
        return ('<a href="/admin/flowr/node/?graph__id__exact=%s"'
            '>Associated Nodes</a>') % obj.id
    self_nodes.allow_tags = True
    self_nodes.short_description = 'Associated Nodes'


@admin.register(Node)
class NodeAdmin(admin.ModelAdmin):
    list_display = ('id', 'self_graph', 'self_parents', 'self_children', 
        'self_is_root', 'data')

    def self_graph(self, obj):
        return '%s' % obj.graph
    self_graph.short_description = 'DCC Graph'

    def self_parents(self, obj):
        parents = [str(p) for p in obj.parents.all()]
        return ', '.join(parents)
    self_parents.short_description = 'Parents'

    def self_children(self, obj):
        children = [str(c) for c in obj.children.all()]
        return ','.join(children)
    self_children.short_description = 'Children'

    def self_is_root(self, obj):
        return obj.is_root()
    self_is_root.short_description = 'Is Root'

# =============================================================================
# Flow Admin Classes
# =============================================================================

@admin.register(RuleSet)
class RuleSetAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'root_rule_label', 'self_flows',
        'self_actions')

    def self_flows(self, obj):
        return ('<a href="/admin/flowr/flow/?rule_set__id__exact=%s"'
            '>Associated Flows</a>') % obj.id
    self_flows.allow_tags = True
    self_flows.short_description = 'Associated Flows'

    def self_actions(self, obj):
        view = reverse('flowr-view-rule-set-tree', args=(obj.id,))
        flow = reverse('flowr-create-flow', args=(obj.id,))
        return ('<a href="%s">View Tree</a>&nbsp;|&nbsp;'
            '<a href="%s">New Flow</a>') % (view, flow)
    self_actions.allow_tags = True
    self_actions.short_description = 'Actions'


@admin.register(Flow)
class FlowAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'rule_set', 'state_graph', 'root_data', 
        'self_actions')
    list_filter = ('rule_set', )

    def self_actions(self, obj):
        flow = reverse('flowr-edit-flow', args=(obj.id,))
        if obj.in_use():
            return '<i>Flow in use</i>'

        return '<a href="%s">Edit Flow</a>' % flow
    self_actions.allow_tags = True
    self_actions.short_description = 'Actions'


@admin.register(FlowNodeData)
class FlowNodeDataAdmin(admin.ModelAdmin):
    list_display = ('id', 'rule_label', 'self_flow', 'self_children')

    def self_children(self, obj):
        kids = [str(node.data.rule_name) for node in obj.node.children.all()]
        return ', '.join(kids)

    def self_flow(self, obj):
        # django admin only detects declared items, this value is set in
        # __init__ so it needs a helper function
        return str(obj.flow)


@admin.register(State)
class StateAdmin(admin.ModelAdmin):
    list_display = ('id', 'flow', 'current_node', 'self_rule', 'self_rule_set')
    list_filter = ('flow', 'flow__rule_set', )

    def self_rule_set(self, obj):
        return str(obj.flow.rule_set)

    def self_rule(self, obj):
        return str(obj.current_node.data.rule.label)



from setuptools import setup, find_packages
from version import VERSION

setup(
    name='django-flowr',
    version=VERSION,
    description='Django based dynamic state machine system',
    long_description=(
        'Most state machine libraries are "static" and require the flow '
        'in the state machine to be definied programmatically.  Flowr '
        'is designed so that you can build state machine flows and '
        'store them in a database.  There are two key concepts: rule '
        'graphs and state machines.  The programmer defines one or more '
        'sets of rules that describe the allowed flow between states, '
        'the user can then use the GUI tools to construct state '
        'machines that follow these rules and store the machines in the '
        'database.  The state machines can then be instantiated for '
        'processing the flow which triggers call-back mechanisms in the '
        'rule objects on entering and leaving a state. '
    ),
    url='https://github.com/cltrudeau/django-flowr',
    author='Christopher Trudeau',
    author_email='ctrudeau+pypi@arsensa.com',
    license='MIT',
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='django,state machine',
    packages=find_packages(),
    install_requires=[
        'Django>=1.8',
    ],
)



# flowr.views.py
import logging, json

from django.contrib.admin.views.decorators import staff_member_required
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, render_to_response
from django.template import RequestContext
from django.template.loader import render_to_string

from flowr.models import RuleSet, Flow, Node

logger = logging.getLogger(__name__)

# ============================================================================
# Flow/Rule Management
# ============================================================================

@staff_member_required
def view_rule_set_tree(request, rule_set_id):
    rule_set = get_object_or_404(RuleSet, id=rule_set_id)
    data = {
        'title':'View: %s' % rule_set.name,
        'graph':rule_set.cytoscape_json(),
        'roots':json.dumps([rule_set.root_rule_name, ] ),
        'return_url': request.META['HTTP_REFERER'],
    }

    t = json.loads(data['graph'])
    print(json.dumps(t, sort_keys=True, indent=4, separators=(',', ': ')))

    return render_to_response('flowr/view_rules.html', data,
        context_instance=RequestContext(request))


@staff_member_required
def create_flow(request, rule_set_id):
    rule_set = get_object_or_404(RuleSet, id=rule_set_id)
    flow = Flow.factory('New Flow', rule_set)
    return edit_flow(request, flow.id)


@staff_member_required
def edit_flow(request, flow_id):
    flow = get_object_or_404(Flow, id=flow_id)

    extra_fields = lambda n: {'label':n.data.rule_name}
    data = {
        'flow':flow,
        'title':'Edit: %s' % flow.name,
        'graph':flow.state_graph.cytoscape_json(extra_fields),
        'roots':json.dumps(['n%s' % flow.state_graph.root.id, ] ),
        'return_url': request.META['HTTP_REFERER'],
    }

    return render_to_response('flowr/edit_flow.html', data,
        context_instance=RequestContext(request))

# ============================================================================
# Node Control
# ============================================================================

@staff_member_required
def node_selected(request, node_id):
    print('got here')
    node = get_object_or_404(Node, id=node_id)
    data = {
        'node':node,
    }

    if node.data.rule.has_edit_screen:
        data['edit_screen'] = node.data.rule.edit_screen(request, node.data)

    result = {
        'html':render_to_string('flowr/node_selected.html', data,
            context_instance=RequestContext(request)),
        'prune': ','.join(['#n%s' % n.id for n in node.prune_list()]),
    }

    return JsonResponse(result)


@staff_member_required
def node_remove(request, node_id):
    node = get_object_or_404(Node, id=node_id)
    result = {
        'selector':'#n%s' % node.id,
    }
    node.remove()

    return JsonResponse(result)


@staff_member_required
def node_prune(request, node_id):
    node = get_object_or_404(Node, id=node_id)
    result = {
        'selector':','.join(['#n%s' % n.id for n in node.prune_list()]),
    }
    node.prune()

    return JsonResponse(result)



from flowr.models import Rule

done_enter = []
done_leave = []

class LocalRule(Rule):
    @classmethod
    def on_enter(cls, state):
        global done_enter
        done_enter.append(cls.__name__)

    @classmethod
    def on_leave(cls, state):
        global done_leave
        done_leave.append(cls.__name__)

    @classmethod
    def display_name(cls):
        return cls.__name__


# Rule Structure:
#
#           A
#          / \
#         B   C
#            / \
#           D   E
#                \
#                 A (loops)

class E(LocalRule):
    children = []

class D(LocalRule):
    children = []

class C(LocalRule):
    children = [D, E]
    multiple_paths = True
    has_edit_screen = True

    @classmethod
    def edit_screen(cls, request, flow_node_data):
        return 'Custom HTML for node: %s' % flow_node_data

class B(LocalRule):
    children = []

class A(LocalRule):
    children = [B, C]

E.children = [A]
